# -*- coding: cp1252 -*-
""" 
ConeLineBuffer.py

Author : fprally
         Many thank to rba for algorythm

Date: 01/25/2011
Update : 03/28/2011 => Replace arcpy.MinimumBoundingGeometry_management process by functions to use it in ArcView license.
                       Convert miles to meters
        

Description : Create cone from line buffer with min buffer value at the start line
              and with max buffer value at the end line.
              
              Warning:
              you must have :
                  1. min value <> max value
                  2. start point of line always have the min value
                  3. end point of line always have the max value
                  4. convert all distances to meters unit
               
Parameters list :
                                Parameters properties
           Name                         Data type          Type      Direction Multiple value Default    Filter    Obtain from
argv[1]   Input Fc                     Feature layer      Required  Input     No              
argv[2]   Input min radius             Field              Required  Input     No                                
argv[3]   Input max radius             Field              Required  Input     No
argv[3]   Output FC                    Feature layer      Required  Input     No  
 ----------------------------------------------------------------------
"""
# Import system modules
import arcpy, os

def GetScratchWorkspace(outPath):
    # Set the scratchWorkspace environment to local file geodatabase
    arcpy.CreateFileGDB_management(outPath, "scratch.gdb", "CURRENT")
    scratchWorkspace = os.path.join(outPath, "scratch.gdb")
    arcpy.env.scratchWorkspace = scratchWorkspace
    return scratchWorkspace

def GetStartPointOfLine(line):
    """
    Return first point of line
    """
    pointStart = arcpy.Point()
    pointStart = line.firstPoint
    return pointStart

def GetEndPointOfLine(line):
    """
    Return end point of line
    """
    pointEnd = arcpy.Point()
    pointEnd = line.lastPoint
    return pointEnd

def BufferPoint(point, dis, scratchWorkspace, outName):
        
    # Process : buffer first point of line
    pointGeometry = arcpy.PointGeometry(point)
    outBufferPoint = scratchWorkspace + "\\" + outName
    if arcpy.Exists(outBufferPoint):
        outBufferPoint = outBufferPoint + "1"
    
    arcpy.Buffer_analysis(pointGeometry, outBufferPoint, dis, "FULL", "ROUND", "NONE", "")
    
    return outBufferPoint


def CreateLineUsingPoint(firstPoint, arcTan, lineLength):
    import math
    """ Get line from first point, angle and length.

    INPUTS:
    first point, end point, length

    OUTPUT:
    line object
    """

    # Process : Create line
    array = arcpy.Array()
    array.add(firstPoint)
    
    # Calculate end point
    endPoint = arcpy.Point()
    endPoint.X = firstPoint.X - math.sin(arcTan)* lineLength
    endPoint.Y = firstPoint.Y - math.cos(arcTan)* lineLength
    array.add(endPoint)
    
    # Create a Polyline object based on the array of points
    polyline = arcpy.Polyline(array)
    array.removeAll()
    return polyline

def ArcTanBetween2points(firstPoint, endPoint):
    import math
    """ Get arcTan between two points.

    INPUTS:
    first point and end point
    
    OUTPUT:
    arcTan
    """
    
    arcTan = math.atan2((firstPoint.X - endPoint.X),(firstPoint.Y - endPoint.Y))
    
    return arcTan

def DistanceBetween2Points(firstPoint, endPoint):
    import math
    """Calculate distance between two points.

    INPUTS:
    first point, end point

    OUTPUT:
    distance
    """
    
    distance = math.sqrt((firstPoint.X - endPoint.X)**2 + (firstPoint.Y - endPoint.Y)**2)

    return distance
    
def GetTangentIntersect(radiusCircleMin, radiusCircleMax, dist2points):
    """ Return intersection point of circle tangent a with the line

    """
    if radiusCircleMax - radiusCircleMin <> 0:
        getTangentIntersect = (radiusCircleMin * dist2points)/ (radiusCircleMax - radiusCircleMin)
    
    return getTangentIntersect

def DeleteFcStoreScratchWorkspace(pathFGDB, wild_card):
    arcpy.env.workspace = pathFGDB
    fcs = arcpy.ListFeatureClasses(wild_card)
    for fc in fcs:
        arcpy.Delete_management(fc)
    return

def GetXYListFromPoint(inPoint):
    # Create a list to store XP of points 
    XYList = []
    # Identify geometry field
    desc = arcpy.Describe(inPoint)
    shapefieldname = desc.ShapeFieldName
    # Loop point
    rows = arcpy.SearchCursor(inPoint)
    for row in rows:
        feat = row.getValue(shapefieldname)
        pnt = feat.getPart()
        # Print x,y coordinates of current point
        XY=[]
        XY.append(pnt.X)
        XY.append(pnt.Y)
        XYList.append(XY)
    
    return XYList

def reorderListXY(inList):
    orderListXY = []
    orderListXY.append(inList[0])
    orderListXY.append(inList[1])
    orderListXY.append(inList[3])
    orderListXY.append(inList[2])
    return orderListXY

def CreatePolygonFromXYList(XYList, outPolyFC):
    
    # Cration d'un tableau pour stocker les points
    array = arcpy.CreateObject("Array")

    # Cration des points en fonction des coord XY de la liste xyList
    for xy in XYList:
        point = arcpy.Point()
        point.X =  xy[0]
        point.Y =  xy[1]
        array.add(point)
        
    # Create a Polygon object based on the array of points
    polygon = arcpy.Polygon(array)
    # Create a copy of the Polygon objects, by using featureList as input to 
    #  the CopyFeatures tool.
    arcpy.CopyFeatures_management(polygon, outPolyFC)
    return

def run_ConeLineBuffer():
    
##    inFC = r"E:\fpr01_CORBEILLE\ConeBuffer\REVERSE_TEST\REVERSE_TEST.shp"
##    inFieldRadiusBufferMin = "MIN_METERS"
##    inFieldRadiusBufferMax = "MAX_METERS"
##    outFC = r"E:\fpr01_CORBEILLE\ConeBuffer\REVERSE_TEST\REVERSE_TEST_CONE.shp"
    
    inFC = arcpy.GetParameterAsText(0)
    inFieldRadiusBufferMin= arcpy.GetParameterAsText(1)
    inFieldRadiusBufferMax= arcpy.GetParameterAsText(2)
    outFC= arcpy.GetParameterAsText(3)
    
    return create_ConeLineBuffer(inFC, inFieldRadiusBufferMin, inFieldRadiusBufferMax, outFC) 

def create_ConeLineBuffer(inFC, inFieldRadiusBufferMin, inFieldRadiusBufferMax, outFC):
    arcpy.AddMessage("*"*10)
    arcpy.AddMessage("Process : Create cone line buffer ...\nrunning ...")
    
    arcpy.OverWriteOutput = True
    
    # Process :  Get Spatial reference
    sr = arcpy.Describe(inFC).spatialReference
    
    # Process : get path of input ...
    outPath = arcpy.Describe(inFC).path
    
    # Process : Get count of features
    result = arcpy.GetCount_management(inFC)
    lineCount = int(result.getOutput(0))
    
    # Process : Set the progressor 
    arcpy.SetProgressor("step", "Create cone line buffer...", 0, lineCount, 1)
    
    # Process : Define the scratch workspace 
    sw = GetScratchWorkspace(outPath)
    arcpy.env.workspace = sw

    # Process : prepare SQL query SHP or GDB
    OID = arcpy.Describe(inFC).OIDFieldName
    delimitedField = arcpy.AddFieldDelimiters(inFC, OID)
    
    # Iterate through the lines in the cursor 
    sRows = arcpy.SearchCursor(inFC)
    for sRow in sRows:
        
        # Process : Get ID and geometry
        index = sRow.getValue(OID)
        shapeLine = sRow.Shape
        
        # Process : Update the progressor label for current feature 
        arcpy.SetProgressorLabel("Create cone for ID " + str(index) + " ...") 
        arcpy.AddMessage("\tCreate cone for ID " + str(index) + " ...")
                         
        # Process : Get field value
        inRadiusBufferMin = sRow.getValue(inFieldRadiusBufferMin)
        inRadiusBufferMax = sRow.getValue(inFieldRadiusBufferMax)

##        # Process : Manage unit
##        if unit.upper() != "DEFAULT":
##            inRadiusBufferMinUnit = "%s %s" % (inRadiusBufferMin, unit)
##            inRadiusBufferMaxUnit = "%s %s" % (inRadiusBufferMax, unit)
        
        # Process : Get start and end point of line
        startPoint = GetStartPointOfLine(shapeLine)
        endPoint = GetEndPointOfLine(shapeLine)
        
        # Process : Create buffer for start/end line
        bufferMin = BufferPoint(startPoint, inRadiusBufferMin, sw, "_bufferMin")
        bufferMax = BufferPoint(endPoint, inRadiusBufferMax, sw, "_bufferMax")
        
        ## Process : Search point a
        dist2points = DistanceBetween2Points(startPoint, endPoint)
        a = GetTangentIntersect(inRadiusBufferMin, inRadiusBufferMax, dist2points)
        radiusCircleMax = (dist2points + a)/2
        
        # Process : Create circle c1 (Max)
        ## Process : Get arctan of line
        valeurArcTan = ArcTanBetween2points(endPoint, startPoint)
        
        ## Process: create line
        lineCircleMax = CreateLineUsingPoint(endPoint, valeurArcTan, radiusCircleMax)
        ## Process : 
        endPtCircleMax = GetEndPointOfLine(lineCircleMax)
        
        ## Process :
        circleMax = BufferPoint(endPtCircleMax, radiusCircleMax, sw, "_circleMax")
        
        # Process : Intersect BufferMaw with CircleMax to get
        # Process : Intersect
        intersectMax = sw + "\\_intersectMax"
        arcpy.Intersect_analysis(bufferMax + ";" + circleMax, intersectMax, "ALL", "", "POINT")
        
        # Process : Create circle c2 (Min)
        radiusCircleMin = a/2
        ## Process : create line
        lineCircleMin = CreateLineUsingPoint(startPoint, valeurArcTan, radiusCircleMin)
        ## Process : 
        endPtCircleMin = GetEndPointOfLine(lineCircleMin)
        ## Process :
        circleMin = BufferPoint(endPtCircleMin, radiusCircleMin, sw, "_circleMin")
        ## Process : Intersect BufferMaw with CircleMax to get
        ## Process : Intersect
        intersectMin = sw + "\\_intersectMin"
        arcpy.Intersect_analysis(bufferMin + ";" + circleMin, intersectMin, "ALL", "", "POINT")
        
        # Process :
        arcpy.env.workspace =sw
        arcpy.Merge_management([intersectMin, intersectMax], "_intersectMerge")
        
        # Process: Multi-parties vers une partie
        singlepartMerge = sw + "\\_singlepartMerge"
        arcpy.MultipartToSinglepart_management("_intersectMerge", singlepartMerge)
        
        # Process : build intermediate geometry bounding with arcview license
        bounding = sw + "\\_bounding"
        XYcoord = GetXYListFromPoint(singlepartMerge)
        orderXYcoord = reorderListXY(XYcoord)
        CreatePolygonFromXYList(orderXYcoord, bounding)
        
        # Process : Merge
        Bounding = sw + "\\_Bounding"
        arcpy.Merge_management([bufferMin, bufferMax, Bounding], "_Bounding_Merge")
        
        # Process : Dissolve
        bounding_Merge = sw + "\\_bounding_Merge"
        arcpy.Dissolve_management(bounding_Merge, "result" + str(index), "", "", "MULTI_PART", "DISSOLVE_LINES")
        
        # Process : delete intermediate fc begin by "_"
        DeleteFcStoreScratchWorkspace(sw, "_*")
        
        # Process : Update the progressor position 
        arcpy.SetProgressorPosition()

    # Process : Merge all result feature classes
    ## Create a value table to hold names of input feature classes for Merge tool.
    vt = arcpy.ValueTable()
    fcs = arcpy.ListFeatureClasses("r*")
    for fc in fcs:
        vt.addRow(fc)

    arcpy.Merge_management(vt, outFC)
    
    # Process : Delete scratch workspace
    arcpy.Delete_management(sw)
    
    arcpy.AddMessage("Process: End")
    arcpy.AddMessage("*"*10)
    arcpy.ResetProgressor()
    
if __name__ == '__main__':
    run_ConeLineBuffer()
    
